from dune.perftool.generation import (include_file,
                                      preamble,
                                      )
from dune.perftool.options import get_option, set_option
from dune.perftool.pdelab.driver import (FEM_name_mangling,
                                         get_cell,
                                         get_dimension,
                                         get_test_element,
                                         get_trial_element,
                                         isDG,
                                         isPk,
                                         isQk,
                                         isQuadrilateral,
                                         isSimplical,
                                         name_initree,
                                         preprocess_leaf_data,
                                         )

from ufl import FiniteElement, MixedElement, TensorElement, VectorElement


@preamble
def typedef_domainfield(name):
    gridt = type_grid()
    return "using {} = {}::ctype;".format(name, gridt)


def type_domainfield():
    typedef_domainfield("DF")
    return "DF"


def basetype_range():
    if get_option('opcounter'):
        from dune.perftool.pdelab.driver.timings import setup_timer
        setup_timer()
        return "oc::OpCounter<double>"
    else:
        return "double"


@preamble
def typedef_range(name):
    return "using {} = {};".format(name, basetype_range())


def type_range():
    name = "RangeType"
    typedef_range(name)
    return name


@preamble
def typedef_grid(name):
    dim = get_dimension()
    if isQuadrilateral(get_trial_element().cell()):
        # For Yasp Grids the jacobi of the transformation is diagonal and constant on each cell
        set_option('diagonal_transformation_matrix', True)
        set_option('constant_transformation_matrix', True)

        range_type = type_range()
        gridt = "Dune::YaspGrid<{0}, Dune::EquidistantCoordinates<{1}, {0}>>".format(dim, range_type)
        include_file("dune/grid/yaspgrid.hh", filetag="driver")
    else:
        if isSimplical(get_trial_element().cell()):
            # gridt = "Dune::UGGrid<{}>".format(dim)
            # include_file("dune/grid/uggrid.hh", filetag="driver")
            gridt = "Dune::ALUGrid<{}, {}, Dune::simplex, Dune::conforming>".format(dim, dim)
            include_file("dune/alugrid/grid.hh", filetag="driver")
        else:
            raise PerftoolCodegenError("Cant match your geometry with a DUNE grid. Please report bug.")
    return "using {} = {};".format(name, gridt)


def type_grid():
    name = "Grid"
    typedef_grid(name)
    return name


@preamble
def define_grid(name):
    include_file("dune/testtools/gridconstruction.hh", filetag="driver")
    ini = name_initree()
    _type = type_grid()
    return ["IniGridFactory<{}> factory({});".format(_type, ini),
            "std::shared_ptr<{}> grid = factory.getGrid();".format(_type)]


def name_grid():
    name = "grid"
    define_grid(name)
    return name


@preamble
def typedef_leafview(name):
    grid = type_grid()
    return "using {} = {}::LeafGridView;".format(name, grid)


def type_leafview():
    name = "GV"
    typedef_leafview(name)
    return name


@preamble
def define_leafview(name):
    _type = type_leafview()
    grid = name_grid()
    return "{} {} = {}->leafGridView();".format(_type, name, grid)


def name_leafview():
    name = "gv"
    define_leafview(name)
    return name


@preamble
def typedef_fem(element, name):
    gv = type_leafview()
    df = type_domainfield()
    r = type_range()
    dim = get_dimension()
    if get_option("blockstructured"):
        include_file("dune/perftool/blockstructured/blockstructuredqkfem.hh", filetag="driver")
        degree = element.degree() * get_option("number_of_blocks")
        return "using {} = Dune::PDELab::BlockstructuredQkLocalFiniteElementMap<{}, {}, {}, {}>;".format(name, gv, df, r, degree)
    if isQk(element):
        include_file("dune/pdelab/finiteelementmap/qkfem.hh", filetag="driver")
        return "using {} = Dune::PDELab::QkLocalFiniteElementMap<{}, {}, {}, {}>;".format(name, gv, df, r, element.degree())
    if isPk(element):
        include_file("dune/pdelab/finiteelementmap/pkfem.hh", filetag="driver")
        return "using {} = Dune::PDELab::PkLocalFiniteElementMap<{}, {}, {}, {}>;".format(name, gv, df, r, element.degree())
    if isDG(element):
        if isQuadrilateral(element.cell()):
            include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="driver")
            # TODO allow switching the basis here!
            return "using {} = Dune::PDELab::QkDGLocalFiniteElementMap<{}, {}, {}, {}>;".format(name, df, r, element.degree(), dim)
        if isSimplical(element.cell()):
            include_file("dune/pdelab/finiteelementmap/opbfem.hh", filetag="driver")
            return "using {} = Dune::PDELab::OPBLocalFiniteElementMap<{}, {}, {}, {}, Dune::GeometryType::simplex>;".format(name, df, r, element.degree(), dim)
        raise NotImplementedError("Geometry type not known in code generation")

    raise NotImplementedError("FEM not implemented in dune-perftool")


def type_fem(element):
    name = "{}_FEM".format(FEM_name_mangling(element).upper())
    typedef_fem(element, name)
    return name


@preamble
def define_fem(element, name):
    femtype = type_fem(element)
    from dune.perftool.pdelab.driver import isDG
    if isDG(element):
        return "{} {};".format(femtype, name)
    else:
        gv = name_leafview()
        return "{} {}({});".format(femtype, name, gv)


def name_fem(element):
    assert isinstance(element, FiniteElement)
    name = "{}_fem".format(FEM_name_mangling(element).lower())
    define_fem(element, name)
    return name


def name_trial_gfs():
    element = get_trial_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    return name_gfs(element, is_dirichlet)


def name_test_gfs():
    element = get_test_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    return name_gfs(element, is_dirichlet)


def name_gfs(element, is_dirichlet, treepath=()):
    if isinstance(element, (VectorElement, TensorElement)):
        subel = element.sub_elements()[0]
        subgfs = name_gfs(subel, is_dirichlet[:subel.value_size()], treepath=treepath + (0,))
        name = "{}_pow{}gfs_{}".format(subgfs,
                                       element.num_sub_elements(),
                                       "_".join(str(t) for t in treepath))
        define_power_gfs(element, is_dirichlet, name, subgfs)
        return name
    if isinstance(element, MixedElement):
        k = 0
        subgfs = []
        for i, subel in enumerate(element.sub_elements()):
            subgfs.append(name_gfs(subel, is_dirichlet[k:k + subel.value_size()], treepath=treepath + (i,)))
            k = k + subel.value_size()
        name = "_".join(subgfs)
        name = "{}_{}".format(name, "_".join(str(t) for t in treepath))
        define_composite_gfs(element, is_dirichlet, name, tuple(subgfs))
        return name
    else:
        assert isinstance(element, FiniteElement)
        name = "{}{}_gfs_{}".format(FEM_name_mangling(element).lower(),
                                    "_dirichlet" if is_dirichlet[0] else "",
                                    "_".join(str(t) for t in treepath))
        define_gfs(element, is_dirichlet, name)
        return name


def type_test_gfs():
    element = get_test_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    return type_gfs(element, is_dirichlet)


def type_trial_gfs():
    element = get_trial_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    return type_gfs(element, is_dirichlet)


def type_gfs(element, is_dirichlet):
    if isinstance(element, (VectorElement, TensorElement)):
        subel = element.sub_elements()[0]
        subgfs = type_gfs(subel, is_dirichlet[:subel.value_size()])
        name = "{}_POW{}GFS".format(subgfs, element.num_sub_elements())
        typedef_power_gfs(element, is_dirichlet, name, subgfs)
        return name
    if isinstance(element, MixedElement):
        k = 0
        subgfs = []
        for subel in element.sub_elements():
            subgfs.append(type_gfs(subel, is_dirichlet[k:k + subel.value_size()]))
            k = k + subel.value_size()
        name = "_".join(subgfs)
        typedef_composite_gfs(element, name, tuple(subgfs))
        return name
    else:
        assert isinstance(element, FiniteElement)
        name = "{}{}_GFS".format(FEM_name_mangling(element).upper(),
                                 "_dirichlet" if is_dirichlet[0] else "",
                                 )
        typedef_gfs(element, is_dirichlet, name)
        return name


@preamble
def define_gfs(element, is_dirichlet, name):
    gfstype = type_gfs(element, is_dirichlet)
    gv = name_leafview()
    fem = name_fem(element)
    return ["{} {}({}, {});".format(gfstype, name, gv, fem),
            "{}.name(\"{}\");".format(name, name)]


@preamble
def define_power_gfs(element, is_dirichlet, name, subgfs):
    gv = name_leafview()
    fem = name_fem(element.sub_elements()[0])
    gfstype = type_gfs(element, is_dirichlet)
    return ["{} {}({}, {});".format(gfstype, name, gv, fem),
            "{}.name(\"{}\");".format(name, name)]


@preamble
def define_composite_gfs(element, is_dirichlet, name, subgfs):
    gfstype = type_gfs(element, is_dirichlet)
    return ["{} {}({});".format(gfstype, name, ", ".join(subgfs)),
            "{}.update();".format(name)]


@preamble
def typedef_gfs(element, is_dirichlet, name):
    vb = type_vectorbackend(element)
    gv = type_leafview()
    fem = type_fem(element)
    cass = type_constraintsassembler(bool(is_dirichlet[0]))
    return "using {} = Dune::PDELab::GridFunctionSpace<{}, {}, {}, {}>;".format(name, gv, fem, cass, vb)


@preamble
def typedef_power_gfs(element, is_dirichlet, name, subgfs):
    include_file("dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh", filetag="driver")
    gv = type_leafview()
    vb = type_vectorbackend(element)
    fem = type_fem(element.sub_elements()[0])
    dim = element.num_sub_elements()
    cass = type_constraintsassembler(bool(is_dirichlet[0]))
    return "using {} = Dune::PDELab::VectorGridFunctionSpace<{}, {}, {}, {}, {}, {}>;".format(name, gv, fem, dim, vb, vb, cass)


@preamble
def typedef_composite_gfs(element, name, subgfs):
    vb = type_vectorbackend(element)
    ot = type_orderingtag()
    args = ", ".join(subgfs)
    return "using {} = Dune::PDELab::CompositeGridFunctionSpace<{}, {}, {}>;".format(name, vb, ot, args)


@preamble
def define_blocksize(name, element):
    from dune.perftool.pdelab.driver import isDG, isQuadrilateral
    assert isDG(element)
    assert isQuadrilateral(element.cell())
    dimension = get_dimension()
    degree = element.degree()
    return "static const int {} = Dune::QkStuff::QkSize<{}, {}>::value;".format(name, degree, dimension)


def name_blocksize(element):
    name = "blocksize"
    define_blocksize(name, element)
    return name


@preamble
def typedef_vectorbackend(name, element):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    if get_option("fastdg"):
        blocksize = name_blocksize(element)
        return "using {} = Dune::PDELab::istl::VectorBackend<Dune::PDELab::istl::Blocking::fixed, {}>;".format(name, blocksize)
    else:
        return "using {} = Dune::PDELab::istl::VectorBackend<Dune::PDELab::istl::Blocking::none, 1>;".format(name)


def type_vectorbackend(element):
    name = "VectorBackend"
    typedef_vectorbackend(name, element)
    return name


def type_orderingtag():
    return "Dune::PDELab::LexicographicOrderingTag"


@preamble
def typedef_dirichlet_constraintsassembler(name):
    include_file("dune/pdelab/constraints/conforming.hh", filetag="driver")
    return "using {} = Dune::PDELab::ConformingDirichletConstraints;".format(name)


@preamble
def typedef_no_constraintsassembler(name):
    return "using {} = Dune::PDELab::NoConstraints;".format(name)


def type_constraintsassembler(is_dirichlet):
    assert isinstance(is_dirichlet, bool)
    if is_dirichlet:
        name = "DirichletConstraintsAssember"
        typedef_dirichlet_constraintsassembler(name)
        return name
    else:
        name = "NoConstraintsAssembler"
        typedef_no_constraintsassembler(name)
        return name


def name_trial_subgfs(treepath):
    if len(treepath) == 0:
        return name_trial_gfs()
    else:
        return name_subgfs(treepath)


def name_subgfs(treepath):
    gfs = name_trial_gfs()
    name = "{}_{}".format(gfs, "_".join(str(t) for t in treepath))
    define_subgfs(name, treepath)
    return name


@preamble
def define_subgfs(name, treepath):
    t = type_subgfs(treepath)
    gfs = name_trial_gfs()
    return "{} {}({});".format(t, name, gfs)


def type_subgfs(tree_path):
    include_file('dune/pdelab/gridfunctionspace/subspace.hh', filetag='driver')
    gfs = type_trial_gfs()
    return "Dune::PDELab::GridFunctionSubSpace<{}, Dune::TypeTree::TreePath<{}> >".format(gfs, ', '.join(str(t) for t in tree_path))
